//-----------------------------------------------
// AiboWeb
// 2003/01/27  Kazuto Sato
//-----------------------------------------------

#include <string.h>
#include <OPENR/OSyslog.h>
#include <OPENR/OPENRAPI.h>
#include <OPENR/core_macro.h>
#include <ant.h>
#include <EndpointTypes.h>
#include <TCPEndpointMsg.h>
#include "AiboWeb.h"
#include "BMP.h"
#include "entry.h"

AiboWeb::AiboWeb ()
{
}

OStatus
AiboWeb::DoInit(const OSystemEvent& event)
{
    OSYSDEBUG(("AiboWeb::DoInit()\n"));
    NEW_ALL_SUBJECT_AND_OBSERVER;
    REGISTER_ALL_ENTRY;
    SET_ALL_READY_AND_NOTIFY_ENTRY;
    return oSUCCESS;
}

OStatus
AiboWeb::DoStart(const OSystemEvent& event)
{
    OSYSDEBUG(("AiboWeb::DoStart()\n"));
    ENABLE_ALL_SUBJECT;
    ASSERT_READY_TO_ALL_OBSERVER;

    ipstackRef = antStackRef("IPStack");

    for (int index = 0; index < AIBOWEB_CONNECTION_MAX; index++) {
        OStatus result = InitTCPConnection(index);
        if (result != oSUCCESS) return oFAIL;
    }

    for (int index = 0; index < AIBOWEB_CONNECTION_MAX; index++)
        Listen(index);

    return oSUCCESS;
}

OStatus
AiboWeb::DoStop(const OSystemEvent& event)
{
    OSYSDEBUG(("AiboWeb::DoStop()\n"));
    DISABLE_ALL_SUBJECT;
    DEASSERT_READY_TO_ALL_OBSERVER;
    return oSUCCESS;
}

OStatus
AiboWeb::DoDestroy(const OSystemEvent& event)
{
    DISABLE_ALL_SUBJECT;
    return oSUCCESS;
}

void AiboWeb::NotifyImage(const ONotifyEvent& event)
{
    OFbkImageVectorData* imageVec = (OFbkImageVectorData*)event.Data(0);

    for (int index = 0; index < AIBOWEB_CONNECTION_MAX; index++) {
        if (connection[index].requested == true) {
            SetSendData(index, imageVec);
            Send(index);
            connection[index].requested = false;
        }
    }
    observer[event.ObsIndex()]->AssertReady(event.SenderID());
}

void AiboWeb::SetSendData(int index, OFbkImageVectorData* imageVec)
{
    BMP bmp;
    char* header = "HTTP/1.0 200 OK\r\n"
                   "Server: AiboWeb\r\n"
                   "Content-Type: image/bmp\r\n\r\n";
    byte* ptr = connection[index].sendData;
    int size;
    
    strcpy((char*)ptr, header);
    ptr += strlen(header);
    bmp.CopyYCrCb2RGB(ptr, imageVec, ofbkimageLAYER_M, &size);
    connection[index].sendSize = strlen(header) + size;
}

OStatus
AiboWeb::Listen(int index)
{
    OSYSDEBUG(("AiboWeb::Listen()\n"));

    if (connection[index].state != CONNECTION_CLOSED) return oFAIL;

    //
    // Create endpoint
    //
    antEnvCreateEndpointMsg tcpCreateMsg(EndpointType_TCP,
                                         AIBOWEB_SEND_BUFFER_SIZE);
    tcpCreateMsg.Call(ipstackRef, sizeof(tcpCreateMsg));
    if (tcpCreateMsg.error != ANT_SUCCESS) {
        OSYSLOG1((osyslogERROR, "%s : %s[%d] antError %d",
                  "AiboWeb::Listen()",
                  "Can't create endpoint",
                  index, tcpCreateMsg.error));
        return oFAIL;
    }
    connection[index].endpoint = tcpCreateMsg.moduleRef;

    //
    // Listen
    //
    TCPEndpointListenMsg listenMsg(connection[index].endpoint,
                                   IP_ADDR_ANY, AIBOWEB_PORT);
    listenMsg.continuation = (void*)index;

    listenMsg.Send(ipstackRef, myOID_,
                   Extra_Entry[entryListenCont], sizeof(listenMsg));
    
    connection[index].state = CONNECTION_LISTENING;

    return oSUCCESS;
}

void
AiboWeb::ListenCont(void* msg)
{
    OSYSDEBUG(("AiboWeb::ListenCont()\n"));

    TCPEndpointListenMsg* listenMsg = (TCPEndpointListenMsg*)msg;
    int index = (int)listenMsg->continuation;

    if (listenMsg->error != TCP_SUCCESS) {
        OSYSLOG1((osyslogERROR, "%s : %s %d",
                  "AiboWeb::ListenCont()",
                  "FAILED. listenMsg->error", listenMsg->error));
        Close(index);
        return;
    }

    connection[index].state = CONNECTION_CONNECTED;

    SetReceiveData(index);
    Receive(index);
}

OStatus
AiboWeb::Send(int index)
{
    OSYSDEBUG(("AiboWeb::Send()\n"));

    if (connection[index].sendSize == 0 ||
        connection[index].state != CONNECTION_CONNECTED) return oFAIL;

    TCPEndpointSendMsg sendMsg(connection[index].endpoint,
                               connection[index].sendData,
                               connection[index].sendSize);
    sendMsg.continuation = (void*)index;

    sendMsg.Send(ipstackRef, myOID_,
                 Extra_Entry[entrySendCont],
                 sizeof(TCPEndpointSendMsg));

    connection[index].state = CONNECTION_SENDING;
    connection[index].sendSize = 0;
    return oSUCCESS;
}

void
AiboWeb::SendCont(void* msg)
{
    OSYSDEBUG(("AiboWeb::SendCont()\n"));

    TCPEndpointSendMsg* sendMsg = (TCPEndpointSendMsg*)msg;
    int index = (int)(sendMsg->continuation);

    if (sendMsg->error != TCP_SUCCESS) {
        OSYSLOG1((osyslogERROR, "%s : %s %d",
                  "AiboWeb::SendCont()",
                  "FAILED. sendMsg->error", sendMsg->error));
        Close(index);
        return;
    }

    connection[index].state = CONNECTION_CONNECTED;
    Close(index);
}

OStatus
AiboWeb::Receive(int index)
{
    OSYSDEBUG(("AiboWeb::Receive()\n"));

    if (connection[index].state != CONNECTION_CONNECTED &&
        connection[index].state != CONNECTION_SENDING) return oFAIL;

    TCPEndpointReceiveMsg receiveMsg(connection[index].endpoint,
                                     connection[index].recvData,
                                     1, connection[index].recvSize);
    receiveMsg.continuation = (void*)index;

    receiveMsg.Send(ipstackRef, myOID_,
                    Extra_Entry[entryReceiveCont], sizeof(receiveMsg));

    return oSUCCESS;
}

void
AiboWeb::ReceiveCont(void* msg)
{
    OSYSDEBUG(("AiboWeb::ReceiveCont()\n"));

    TCPEndpointReceiveMsg* receiveMsg = (TCPEndpointReceiveMsg*)msg;
    int index = (int)(receiveMsg->continuation);

    if (receiveMsg->error != TCP_SUCCESS) {
        OSYSLOG1((osyslogERROR, "%s : %s %d",
                  "AiboWeb::ReceiveCont()",
                  "FAILED. receiveMsg->error", receiveMsg->error));
        Close(index);
        return;
    }

    connection[index].recvData[receiveMsg->sizeMax] = 0;
    OSYSPRINT(("* Requested *\n"));
    printRecvData((char*)connection[index].recvData);

    connection[index].requested = true;
}

OStatus
AiboWeb::Close(int index)
{
    OSYSDEBUG(("AiboWeb::Close()\n"));

    if (connection[index].state == CONNECTION_CLOSED ||
        connection[index].state == CONNECTION_CLOSING) return oFAIL;

    TCPEndpointCloseMsg closeMsg(connection[index].endpoint);
    closeMsg.continuation = (void*)index;

    closeMsg.Send(ipstackRef, myOID_,
                  Extra_Entry[entryCloseCont], sizeof(closeMsg));

    connection[index].state = CONNECTION_CLOSING;

    return oSUCCESS;
}

void
AiboWeb::CloseCont(void* msg)
{
    OSYSDEBUG(("AiboWeb::CloseCont()\n"));
    
    TCPEndpointCloseMsg* closeMsg = (TCPEndpointCloseMsg*)msg;
    int index = (int)(closeMsg->continuation);

    connection[index].state = CONNECTION_CLOSED;
    Listen(index);
}

OStatus
AiboWeb::InitTCPConnection(int index)
{
    OSYSDEBUG(("AiboWeb::InitTCPConnection()\n"));

    connection[index].state = CONNECTION_CLOSED;

    // 
    // Allocate send buffer
    //
    antEnvCreateSharedBufferMsg sendBufferMsg(AIBOWEB_SEND_BUFFER_SIZE);

    sendBufferMsg.Call(ipstackRef, sizeof(sendBufferMsg));
    if (sendBufferMsg.error != ANT_SUCCESS) {
        OSYSLOG1((osyslogERROR, "%s : %s[%d] antError %d",
                  "AiboWeb::InitTCPConnection()",
                  "Can't allocate send buffer",
                  index, sendBufferMsg.error));
        return oFAIL;
    }

    connection[index].sendBuffer = sendBufferMsg.buffer;
    connection[index].sendBuffer.Map();
    connection[index].sendData
        = (byte*)(connection[index].sendBuffer.GetAddress());

    //
    // Allocate receive buffer
    //
    antEnvCreateSharedBufferMsg recvBufferMsg(AIBOWEB_BUFFER_SIZE);

    recvBufferMsg.Call(ipstackRef, sizeof(recvBufferMsg));
    if (recvBufferMsg.error != ANT_SUCCESS) {
        OSYSLOG1((osyslogERROR, "%s : %s[%d] antError %d",
                  "AiboWeb::InitTCPConnection()",
                  "Can't allocate receive buffer",
                  index, recvBufferMsg.error));
        return oFAIL;
    }

    connection[index].recvBuffer = recvBufferMsg.buffer;
    connection[index].recvBuffer.Map();
    connection[index].recvData
        = (byte*)(connection[index].recvBuffer.GetAddress());

    connection[index].requested = false;

    return oSUCCESS;
}

void
AiboWeb::SetReceiveData(int index)
{
    connection[index].recvSize = AIBOWEB_BUFFER_SIZE;
}

//-----------------------------------------------
// o
//-----------------------------------------------
void AiboWeb::printRecvData(const char* src)
{
    char buf[201];
    const char *p = src;
    int len = strlen(src);
    int nokori = len;
    
    while(nokori > 0) {
        int i = 200;
        if(i > nokori) i = nokori;
        strncpy(buf, p, i); buf[i] = 0;
        OSYSPRINT((buf));
        nokori -= i;
        p += i;
    }
}
